#!/usr/bin/python
# -*- coding: utf-8 -*-

import time
from os import path
from httplib2 import Http
import urllib
from math import *
from lxml import etree
from PIL import Image, ImageDraw
from StringIO import StringIO
from geohash import encode
from geographiclib.geodesic import Geodesic

# EC Equator lenght
EC = 40075016.686 # meter
# ER Earth radius
ER = 6372798.2 # meter
# Availible zoom levels in OSM
ZOOMRANGE = range(1, 18)
# tile size
TS = 256

DB_NS="http://docbook.org/ns/docbook"
DB = "{%s}" % DB_NS
XI_NS="http://www.w3.org/2001/XInclude"
XI = "{%s}" % XI_NS
XLINK_NS="http://www.w3.org/1999/xlink"
XLINK = "{%s}" % XLINK_NS
HTML_NS="http://www.w3.org/1999/xhtml"
HTML = "{%s}" % HTML_NS
NSMAP = {None : DB_NS,
         'xlink' : XLINK_NS}


h = Http(".cache")

class Coord(object):
    def __init__(self, lat, lon):
        self.latitude  = float(lat)
        self.longitude = float(lon)
        self.image = None

    def osmlink(self):
        return "http://www.openstreetmap.org/?mlat=%s&mlon=%s&zoom=18&layers=M"\
            % (self.latitude,self.longitude)

    def dms(self):
        ns = self.latitude
        ew = self.longitude
        mnt,sec = divmod(ns*3600,60)
        deg,mnt = divmod(mnt,60)
        out = u'''%d°%2d'%5.2f"%s''' % ( deg,mnt,sec,'N')
        mnt,sec = divmod(ew*3600,60)
        deg,mnt = divmod(mnt,60)
        out +=  u''' %d°%2d'%05.2f"%s''' % ( deg,mnt,sec,'E')
        return out

    def lontile(lon, zoom):
        tile = ((lon + 180) / 360) * (2**zoom)
        return tile

    def lattile(lat, zoom):
        tile = (1 - log(tan(lat * pi/180) + 1 / cos(lat* pi/180)) / pi) /2 * 2**zoom
    #    tile = (1-log(tan(radians(lat))+1/cos(radians(lat)))/pi)/2*2**zoom
        return tile

    def coordtile(coord, zoom):
        x = lontile(coord[1],zoom)
        y = lattile(coord[0],zoom)
        return (y,x)

    def tile(x,y,z):
        return 'http://tile.openstreetmap.org/%s/%s/%s.png' % (z,x,y)

    def link(coord,zoom):
        (x, y) = coordtile(coord,zoom)
        x = int(floor(x))
        y = int(floor(y))
        return tile(x,y,zoom)

    def offset(coord, zoom):
        (x, y) = coordtile(coord,zoom)
        xo = int(floor((x-floor(x))*TS))
        yo = int(floor((y-floor(y))*TS))
        return (xo, yo)

    def distance(self, point):
        res = Geodesic.WGS84.Inverse(self.latitude, self.longitude,
                                     point.latitude, point.longitude)
        return res['s12']

    def png(self,zoom=15,size=(TS,TS)):
        filename = encode(self.latitude, self.longitude)+'.png'
        if path.isfile(filename):
            if path.getctime(filename) > time.time() - 60*60*24*2:
                return
        im = Image.new("RGB", size, None)
        center = (size[0]/2, size[1]/2)
        nst = ceil(center[0]-TS/2)/TS
        ewt = ceil(center[1]-TS/2)/TS
        x, y = offset(co,zoom)
        (ns, ew) = coordtile(co,zoom)
        if x < TS/2:
            e = 1
            w = 0
        if x > TS/2:
            e = 0
            w = 1
        if y < TS/2:
            n = 1
            s = 0
        if y > TS/2:
            n = 0
            s = 1
        ul = (ns-nst-n, ew-ewt-e)
        lr = (ns+nst+s, ew+ewt+w)
        lattiles = range(int(ul[0]),int(lr[0])+1)
        lontiles =  range(int(ul[1]),int(lr[1])+1)
#        url = link(c,zoom)
        size = (len(lontiles)*TS,len(lattiles)*TS)
        grid = Image.new("RGB", size, None)
        for yi, y in enumerate(lattiles):
            for xi, x in enumerate(lontiles):
                url = tile(x,y,zoom)
                time.sleep(1)
                request, content = h.request(url)
                img = Image.open(StringIO(content))
                box = (xi*TS, yi*TS)
                grid.paste(img, box)
        t = coordtile(co,zoom)
        o = offset(co,zoom)
        yp = [i for i,x in enumerate(lattiles) if x == int(t[0])][0]*TS+o[1]
        xp = [i for i,x in enumerate(lontiles) if x == int(t[1])][0]*TS+o[0]
        mark(grid, (xp,yp))
        gridc = grid.crop((xp-TS/2,yp-TS/2,xp+TS/2,yp+TS/2))
        gridc.save(filename)

    def db_xml(self):
        uri = etree.Element(DB+'uri',nsmap=NSMAP)
        ln  = etree.SubElement(uri,  DB+'link')
        ln.set(XLINK+'href',self.osmlink())
        imo  = etree.SubElement(ln,  DB+'inlinemediaobject')
        io   = etree.SubElement(imo, DB+'imageobject', condition="web")
        idat = etree.SubElement(io , DB+'imagedata',
                                fileref=encode(self.latitude, self.longitude)+'.png',
                                format='PNG')
        to   = etree.SubElement(imo, DB+'textobject')
        ph   = etree.SubElement(to,  DB+'phrase')
        ph.text = "geo:"+str(self.latitude)+","+str(self.longitude)
        para = etree.SubElement(ln,  DB+'para')
        para.text = self.dms()
        return uri

class Address(object):
    """Address object to contain everything known about an address"""
    def __init__(self,address):
        self._address_string = address
        self._root = etree.Element(DB+'address',nsmap=NSMAP)
        self._coord = None

    def geocode(self,country=None):
        params = { 'q': self._address_string,
                   'addressdetails': 1,
                   'limit': 1,
                   'format': 'xml',
                   'polygon': 0 }

        if country:
            params['countrycodes'] = country

        base_url = 'http://nominatim.openstreetmap.org/search?%s'
        url = base_url % urllib.urlencode(params)
        resp, content = h.request(url)
        root = etree.fromstring(content)
        place = root.find("place")
        if place is not None:
            print (etree.tostring(root, pretty_print=True))
            self._coord=Coord(place.get("lat"),place.get("lon"))
            return 1
        else:
            print resp
            print content
            return 0

    def db_xml(self):
        return self._coord.db_xml()


def distance(p1, p2):
    res = Geodesic.WGS84.Inverse(p1[0], p1[1], p2[0], p2[1])
    return res['s12']

def mark(image, coord):
    draw = ImageDraw.Draw(image)
    x, y = coord
    r = 5
    bbox = (x-r, y-r, x+r, y+r)
    draw.ellipse(bbox, outline="red")

def mapimages(coords, zoom=15,size=(TS,TS)):
        minlat = 1000
        maxlat = 0
        minlon = 1000
        maxlon = 0
        for c in coords:
            minlat = min(minlat,c[0])
            maxlat = max(maxlat,c[0])
            minlon = min(minlon,c[1])
            maxlon = max(maxlon,c[1])
        # Find minimal bounding box and expand it 5%
        hyp = distance((maxlat,minlon),(minlat,maxlon))
        hyp = hyp*0.05
        uld =  Geodesic.WGS84.Direct(maxlat, minlon, 315, hyp)
        urd =  Geodesic.WGS84.Direct(maxlat, maxlon,  45, hyp)
        lld =  Geodesic.WGS84.Direct(minlat, minlon, 225, hyp)
        lrd =  Geodesic.WGS84.Direct(minlat, maxlon, 135, hyp)

        ul = (uld['lat2'],uld['lon2'])
        ur = (urd['lat2'],urd['lon2'])
        ll = (lld['lat2'],lld['lon2'])
        lr = (lrd['lat2'],lrd['lon2'])
        top = distance(ul,ur)
        bottom = distance(ll,lr)
        left = distance(ul,ll)
        right = distance(ur,lr)
#    m_per_pix = EC*abs(cos(lat))/2**(zoomlevel+8)
#    m_per_pix = ER*2*pi*abs(cos(lat))/2**(zoomlevel+8)
        for zoom in range(1,18):
            t = 2**(zoom)*TS/(ER*2*pi*abs(cos(ul[0])))
            toppix = t*top
            leftpix = t*left
            b = 2**(zoom)*TS/(ER*2*pi*abs(cos(ll[0])))
            bottompix = b*bottom
            rightpix  = b*right
#            print "Zoom: %s : %s %s %s %s" % (zoom, toppix, bottompix, leftpix, rightpix)
            if max(toppix,bottompix,leftpix,rightpix) > TS*4:
                break
#        print "Zoom to use : %s" % (zoom)
        ult = coordtile(ul,zoom)
        lrt = coordtile(lr,zoom)
        lattiles = range(int(ult[0]),int(lrt[0])+1)
        lontiles =  range(int(ult[1]),int(lrt[1])+1)
        size = (len(lontiles)*TS,len(lattiles)*TS)
        grid = Image.new("RGB", size, None)
        img = []
        for yi, y in enumerate(lattiles):
            for xi, x in enumerate(lontiles):
                url = tile(x,y,zoom)
#                print url
                time.sleep(1)
                request, content = h.request(url)
                img = Image.open(StringIO(content))
                box = (xi*TS, yi*TS)
                grid.paste(img, box)
        for c in coords:
            t = coordtile(c,zoom)
            o = offset(c,zoom)
            yp = [i for i,x in enumerate(lattiles) if x == int(t[0])][0]*TS+o[1]
            xp = [i for i,x in enumerate(lontiles) if x == int(t[1])][0]*TS+o[0]
            mark(grid, (xp,yp))

        t = coordtile(ul,zoom)
        o = offset(ul,zoom)
        yp = [i for i,x in enumerate(lattiles) if x == int(t[0])][0]*TS+o[1]
        xp = [i for i,x in enumerate(lontiles) if x == int(t[1])][0]*TS+o[0]
        t = coordtile(lr,zoom)
        o = offset(lr,zoom)
        ypl = [i for i,x in enumerate(lattiles) if x == int(t[0])][0]*TS+o[1]
        xpl = [i for i,x in enumerate(lontiles) if x == int(t[1])][0]*TS+o[0]
        gridc = grid.crop((xp,yp,xpl,ypl))
        gridc.show()
#        gridc.save("cap-un.png")
